/* DirLister

		Copyright (c) 2002-2003 Olivier MILLA
			
		Designed & Programmed by Olivier MILLA (Methedras) 
								 methedras@online.fr
		  			  			 
	Source Code Provided under the Artistic License		  
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <Alert.h>
#include <AppFileInfo.h>
#include <Directory.h>
#include <Font.h>
#include <Path.h>
#include <Rect.h>
#include <String.h>
#include <SymLink.h>
#include <TrackerAddon.h>

#include "CenterOnScreen.h"
#include "DirectoryLister.h"

#define B_GREY 		216,216,216
#define GO_M   		'gobm'
#define DATA_SPACE	20

#define DEBUG		0

//The GUI
DirListerWindow::DirListerWindow() : BWindow(CenterRectOnScreen(BRect(0,0,450,180)),"DirLister",B_TITLED_WINDOW,B_NOT_ANCHORED_ON_ACTIVATE | B_NOT_ZOOMABLE | B_NOT_RESIZABLE,B_CURRENT_WORKSPACE)
{
	BView *view = new BView(Bounds(),"view",B_FOLLOW_ALL_SIDES,B_WILL_DRAW);
	
	view->SetViewColor(B_GREY);
		
	fsubDirectories = new BCheckBox(BRect(10,10,250,20),"subdirs","List subdirectories",new BMessage());
	fsymlinks 		= new BCheckBox(BRect(10,30,250,40),"symlinks","Include symlinks (Adds \"(S)\" to SymLinks)",new BMessage());	
	fshow			= new BCheckBox(BRect(10,50,250,60),"show","Show output when done",new BMessage());
	
	fsubDirectories->SetValue(1);
	fsymlinks->SetValue(0);	
	fshow->SetValue(1);
		
	view->AddChild(fsubDirectories);
	view->AddChild(fsymlinks);	
	view->AddChild(fshow);
		
	foutputPath = new BTextControl(BRect(10,75,250,85),"output_path","Output :",NULL,new BMessage());
	foutputPath->SetDivider(foutputPath->Divider() - 80);
	foutputPath->SetAlignment(B_ALIGN_RIGHT,B_ALIGN_LEFT);
	view->AddChild(foutputPath);
	
	fgo = new BButton(BRect(390,120,445,170),"go","Go !!",new BMessage(GO_M));
	fgo->MakeDefault(true);
	view->AddChild(fgo);
	
			
	fdata = new BBox(BRect(255,10,440,110),"data");
	fdata->SetLabel(" Additional Datas displayed");
	view->AddChild(fdata);	
	
	fsize = new BCheckBox(BRect(10,20,120,30),"size","Size",new BMessage());
	fdata->AddChild(fsize);
	
	fcreatetime = new BCheckBox(BRect(10,40,140,50),"createtime","Creation time (CT)",new BMessage());
	fdata->AddChild(fcreatetime);
	
	fmodifiedtime = new BCheckBox(BRect(10,60,140,70),"modifiedtime","Last modified time (MT)",new BMessage());
	fdata->AddChild(fmodifiedtime);
	
	DirListerIconView *icon = new DirListerIconView();
	view->AddChild(icon);
	
/*	fstatus = new BStringView(BRect(60,120,290,135),"StringView_1","Status: Configurating...");
	BFont boldfont(be_bold_font);
	fstatus->SetFont(&boldfont);	
	view->AddChild(fstatus); */
	
	BStringView *text = new BStringView(BRect(10,155,310,165),"StringView_2","DirLister 1.3 ©2002-2003 Olivier MILLA (methedras@online.fr)");
	view->AddChild(text);
	
	AddChild(view);
}

//The Icon
DirListerIconView::DirListerIconView() : BView(BRect(10,110,330,142),"icon",B_FOLLOW_ALL_SIDES,B_WILL_DRAW)
{
	SetViewColor(216,216,216);
	ficon = new BBitmap(BRect(0,0,31,31),B_CMAP8);
}

DirListerIconView::~DirListerIconView()
{
	delete ficon;
}

void DirListerIconView::Draw(BRect updateRect)
{
	BFile file("/boot/home/config/add-ons/Tracker/DirLister",B_READ_ONLY);
	BAppFileInfo fileinfo(&file);
	
	if (fileinfo.InitCheck() == B_OK);
		if(fileinfo.GetIcon(ficon, B_LARGE_ICON) != B_OK)
			fileinfo.GetTrackerIcon(ficon, B_LARGE_ICON);

	if (ficon)
		DrawBitmap(ficon,BPoint(0,0));
	
}

/*This func returns the 'level' of a file comparing to the selected 'root' directory
 for instance, a file into the selected directory has the level 0,
 one into a first sub-directory has level 1, etc...*/
int32 LevelOf(const char *path,int32 rootDirPathCountChars)
{
	BString pathS = path;
	
	pathS.Remove(0,rootDirPathCountChars);
	
	char *pathC = pathS.LockBuffer(pathS.Length() + 1);
	int32 level = 0;
	for(int32 i=0;i<pathS.CountChars();i++)
		if (pathC[i] == '/')
			level++;
	
	pathS.UnlockBuffer(-1);		
	
	return level-1;		
}

/*When reached the end of a directory, the directory is 'off' which means that
 we should no longer draw a '|' at his level. So, this func says whether the parent
 directories of a file are 'off' or not*/ 
bool IsLevelOff(int32 level,int32 LevelsOff[],int32 numberOfLevelsOff)
{
	for (int32 i=0;i<numberOfLevelsOff;i++){
		if (level == LevelsOff[i])
			return true;
	}
	return false;		
				
}

/*This recursive func lists the files. The output is a BString which contains 
  the listing which you will see into the end textfile whith some things:
 	_ When a file is added, at it's end of the line, '////' is put, 
 	  the first two '/' will be replcaed by the file size by AddDatas(), just like
 	  the two next will be changed for the file's creation time and modified time*/
BString ListFiles(BDirectory *directory,int32 *longestLine,bool subdirs,bool symlinks,int32 rootDirPathCountChars,BMessage levelsOff,bool slash)
{
	//Gets the levels which are off and put them into an array for use
	int32 numberOfLevelsOff;
	levelsOff.FindInt32("number_of_levels_off",&numberOfLevelsOff);
	int32 *LevelsOff = new int32[numberOfLevelsOff];
	for (int32 i=0;i<numberOfLevelsOff;i++)
		levelsOff.FindInt32("levels_off",i,&LevelsOff[i]);				
	
	//The file's level
	int32 level=0;
	
	/*All the text that will be added by this passage into ListFiles() 
	(remember, this func is recursive.)*/
	BString here;
				
	if (directory->InitCheck() == B_OK){
		BEntry entry;
		int32 counter = directory->CountEntries();
		while (directory->GetNextEntry(&entry) == B_OK){
				BString line;
			if (entry.IsSymLink() && symlinks == true || entry.IsSymLink() == false){
				
				char name[B_FILE_NAME_LENGTH];
				entry.GetName(name);
				BPath path;
				entry.GetPath(&path);
				
				//Gets the level of the current file
				level = LevelOf(path.Path(),rootDirPathCountChars);									
				
				if (level > 0){
					if (counter == 1){ //If this is the last file of the parent directory
						for (int32 i=0;i<level-1;i++){
							if (IsLevelOff(i,LevelsOff,numberOfLevelsOff))
								line << "\t";
							else {
								line << "\t";
								line << "|";							
							}
						}
						line << "\t";
						line << "`" ;												
					}
					else {
						for (int32 i=0;i<level-1;i++){
							if (IsLevelOff(i,LevelsOff,numberOfLevelsOff))
								line << "\t";
							else {
								line << "\t";
								line << "|" ;
							}	
						}
						line << "\t" ;
						line << "|" ;										
					}
					line << "--" ;									
				}				
				line << name ;
				//Adds '(S)' if the file is a symlink
				if (entry.IsSymLink()) 
					line << " (S)" ;
				
				//Adds the '////' needed for ListDatas()
				if (slash)	
					line << "////" ;					
					
				line << "\n" ;
				
				//See if the line is the longest or not.
				if (line.CountChars() > *longestLine)
					*longestLine = line.CountChars();
			}
			if (entry.IsDirectory() == true && subdirs == true){
				BDirectory nextdir(&entry);
				if (counter == 1){ //If this directory is the latest entry of parent's directory
					levelsOff.RemoveData("number_of_levels_off");
					numberOfLevelsOff++;	
					levelsOff.AddInt32("number_of_levels_off",numberOfLevelsOff);
					levelsOff.AddInt32("levels_off",level-1);
						
					line << ListFiles(&nextdir,longestLine,subdirs,symlinks,rootDirPathCountChars,levelsOff,slash);	
				}
				else
					line << ListFiles(&nextdir,longestLine,subdirs,symlinks,rootDirPathCountChars,levelsOff,slash);
			}	
			here << line;
			counter--;
		}		
	}
	else {
		//Manage the error
	}
	delete LevelsOff;
	return here;
}

//Insert spaces to line the datas, variable's name talk by themselves.
void MoveBy(BString *output,int32 longestLine)
{
	int32 i=0;
	int32 beginning_of_line=0, end_of_line=0, length_of_line=0;
	
	//for the first line
	if ((beginning_of_line = (output->FindLast("\n",output->FindFirst("//")))+1) == B_ERROR)
		beginning_of_line = 0;
		
	end_of_line = output->FindFirst("//");
	
	length_of_line = end_of_line - beginning_of_line;
		
	for (i=0; i<(longestLine-length_of_line); i++)
		output->Insert(" ",output->FindFirst("//"));

#if(DEBUG == 1)	
	printf("MoveBy::length_of_line =%d\n",length_of_line);	
	printf("MoveBy::i =%d\n",i);	
#endif
}

//Recursively Lists file's datas and put them at the place of the '/' (see ListFiles() comment)
void ListDatas(BDirectory *directory,BString *output,int32 longestLine, bool subdirs,bool symlinks,bool size,bool cretime,bool modtime)
{				
	if (directory->InitCheck() == B_OK){
		BEntry entry;
		while (directory->GetNextEntry(&entry) == B_OK){
			if (entry.IsSymLink() && symlinks == true || entry.IsSymLink() == false){

				MoveBy(output,longestLine);					
				if (entry.IsSymLink()) 
					output->ReplaceFirst("//","SYMLINK");
				else if (size == true){
					off_t size;
					BString tochar;					
					entry.GetSize(&size);
					if (size > 1000000){
						size /=1000000;
						tochar << size;
						tochar << " Mo";
					}			
					else if (size > 1000){
						size /= 1000;
						tochar << size;
						tochar << " Ko";
					}	
					else {
						tochar << size;
						tochar << " octets";
					}														
					output->ReplaceFirst("//",tochar.String());		
				}
				else if (size == false)
					output->RemoveFirst("//");
					
				BString timing = " ";
				if (cretime) {
					time_t create_time;
  					entry.GetCreationTime(&create_time);
					timing << "\t\t\tCT:" <<ctime(&create_time);
				}
				if (modtime){
					time_t modif_time;
  					entry.GetCreationTime(&modif_time);
					timing << "\t\t\tMT:" << ctime(&modif_time);					
				}
				timing.RemoveAll("\n"); /*I don't know where these "\n" come from but
										they are here and should not. */
				output->ReplaceFirst("//",timing.String());				
			}
			if (entry.IsDirectory() == true && subdirs == true){
				BDirectory nextdir(&entry);
				ListDatas(&nextdir,output,longestLine,subdirs,symlinks,size,cretime,modtime);
			}	
		}		
	}
	else {
		//Manage the error
	}
}

/*The MessageReceived() of the main window: it only manages 'GO_M' message
  which launchs the listing.*/
void DirListerWindow::MessageReceived(BMessage *message)
{
	switch (message->what){
		case GO_M:{		
			//fstatus->SetText("Status: Processing..."); //Works too late
			//Should wait for update to be done
			
			BDirectory rootDir(&fentryReceived);
			
			BPath path;
			fentryReceived.GetPath(&path);
			BString rootPath;
			rootPath << path.Path();			
			
			FILE *output = fopen(foutputPath->Text(),"w");	
			
			bool subdirs;
			if (fsubDirectories->Value() > 0)
				subdirs = true;
			else
				subdirs = false;
				 
			bool symlinks;
			if (fsymlinks->Value() > 0)
				symlinks = true;
			else
				symlinks = false;
				
			bool size;
			if (fsize->Value() > 0)
				size = true;
			else
				size = false;
			
			bool cretime;
			if (fcreatetime->Value() > 0)
				cretime = true;
			else
				cretime = false;
							
			bool modtime;
			if (fmodifiedtime->Value() > 0)
				modtime = true;
			else
				modtime = false;
					
			BMessage levelsOff;
			levelsOff.AddInt32("number_of_levels_off",0);				
			BString text;
			int32 longestLine=0;
			
			bool slash = false;	
			if (size || cretime || modtime)
				slash = true;
										
			text << ListFiles(&rootDir,&longestLine,subdirs,symlinks,rootPath.CountChars(),levelsOff,slash);			
#if(DEBUG == 1)
			printf("Longest Line =%d\n",longestLine);
#endif			
			if (size || cretime || modtime){
				longestLine+=20; //Adds spaces between the end of the line and datas
			
				//Reinitialize to root directory
				rootDir = BDirectory(&fentryReceived);
				ListDatas(&rootDir,&text,longestLine,subdirs,symlinks,size,cretime,modtime);
			}
			
			fprintf(output,text.String());
			
			fclose(output);	
			
			if (fshow->Value() > 0){
				BString sys = "/boot/apps/StyledEdit ";
				sys << "\"" << foutputPath->Text() << "\"";
				sys << "&";
				system(sys.String());
			}
			
			Quit();		
			break;
		}
		default:{
			BWindow::MessageReceived(message);
			break;
		}	
	}
}

/*This func change a selected symlink's entry into the pointed directory's entry.
 Note: this func is recursive so it can move from a symlink to another if the
       selected symlink points onto another symlink, etc...
 It returns true if everything went fine, false otherwise.*/
bool Handle_Symlink(BEntry *symlink)
{
	BPath path;
	BDirectory parent_dir;
	BSymLink sl(symlink);
				
	symlink->GetParent(&parent_dir);
				
	sl.MakeLinkedPath(&parent_dir,&path);
	*symlink = BEntry(path.Path());
	if (symlink->IsDirectory())
		return true;
	else if (symlink->IsSymLink())
		return Handle_Symlink(symlink);
	else{
		BAlert *dAlert = new BAlert("Directory","You must select a directory","Oups...",NULL,NULL,B_WIDTH_FROM_WIDEST,B_INFO_ALERT);
		dAlert->Go();		
		return false;
	}				
}

//The func called by the Tracker
void process_refs(entry_ref dir_ref, BMessage *msg, void *)
{	
	entry_ref file_ref;	
	
	//Be sure only one directory has been selected
	if (msg->FindRef("refs",1,&file_ref) == B_NO_ERROR){
		BAlert *odAlert = new BAlert("One Directory","You must select only one directory","Oups...",NULL,NULL,B_WIDTH_FROM_WIDEST,B_INFO_ALERT);
		odAlert->Go();	
	}
	else if (msg->FindRef("refs",&file_ref) == B_NO_ERROR){
		DirListerWindow *window = new DirListerWindow();
				
		BEntry entry(&file_ref);
		
		if (entry.IsDirectory() || entry.IsSymLink()){
		
			//Take the 'pointed' directory by the symlink
			if (entry.IsSymLink())
				if (Handle_Symlink(&entry) == false)
					return;
			
			window->fentryReceived = entry; //Saves the selected entry for further usages
				
			char name[B_FILE_NAME_LENGTH];
			entry.GetName(name);

			time_t Time;
			time(&Time);
			BString STime = ctime(&Time);
		
			BString outputPath = "/boot/home/";	
			outputPath << "listing.[";
			outputPath << name;
			outputPath << "]@";						
			outputPath << STime;
					
			window->Lock();
			window->foutputPath->SetText(outputPath.String());
			window->Unlock();
			
			window->Show(); 

			int32 result;
			wait_for_thread(window->Thread(),&result);	/*Wait for the Thread of the window 
													to finish otherwise, the add-on 
													is unloaded	as we are at the end 
													of process_refs() and therefore,
													the window, etc... do not exist
														anymore */		
		}
		else {
			BAlert *dAlert = new BAlert("Directory","You must select a directory","Oups...",NULL,NULL,B_WIDTH_FROM_WIDEST,B_INFO_ALERT);
			dAlert->Go();				
		}	
	}		
}

/*The application: needed to tell the user of his mistake if
  he tries to Launch DirLister by double clicking it.*/
DirLister::DirLister() : BApplication("application/x-vnd.DirLister")
{
	BAlert *alert = new BAlert("Alert","This application is a Tracker Add-on, you can't launch it this way.","Oups...",NULL,NULL,B_WIDTH_FROM_WIDEST,B_INFO_ALERT);
	alert->Go();
	PostMessage(B_QUIT_REQUESTED);
}

DirLister::~DirLister()
{}

int main()
{
	DirLister *app = new DirLister();
	app->Run();
	delete app;
	
	return B_NO_ERROR;
}

